import re
import urllib.parse
from collections import OrderedDict
import singer
from tap_google_sheets.streams import STREAMS

LOGGER = singer.get_logger()

# Reference:
# https://github.com/singer-io/getting-started/blob/master/docs/DISCOVERY_MODE.md#Metadata

# Convert column index to column letter
def colnum_string(num):
    string = ""
    while num > 0:
        num, remainder = divmod(num - 1, 26)
        string = chr(65 + remainder) + string
    return string


def pad_default_effective_values(headers, first_values):
    for i in range(len(headers) - len(first_values)):
        first_values.append(OrderedDict())


# Create sheet_metadata_json with columns from sheet
def get_sheet_schema_columns(sheet):
    sheet_title = sheet.get('properties', {}).get('title')
    sheet_json_schema = OrderedDict()
    data = next(iter(sheet.get('data', [])), {})
    row_data = data.get('rowData', [])
    if row_data == [] or len(row_data) == 1:
        # Empty sheet or empty first row, SKIP
        LOGGER.info('SKIPPING Empty Sheet: {}'.format(sheet_title))
        return None, None

    # spreadsheet is an OrderedDict, with orderd sheets and rows in the repsonse
    headers = row_data[0].get('values', [])
    first_values = row_data[1].get('values', [])
    # Pad first row values with default if null
    if len(first_values) < len(headers):
        pad_default_effective_values(headers, first_values)

    sheet_json_schema = {
        'type': 'object',
        'additionalProperties': False,
        'properties': {
            '__sdc_spreadsheet_id': {
                'type': ['null', 'string']
            },
            '__sdc_sheet_id': {
                'type': ['null', 'integer']
            },
            '__sdc_row': {
                'type': ['null', 'integer']
            }
        }
    }

    header_list = [] # used for checking uniqueness
    columns = []
    prior_header = None
    i = 0
    skipped = 0

    # if no headers are present, log the message that sheet is skipped
    if not headers:
        LOGGER.warning('SKIPPING THE SHEET AS HEADERS ROW IS EMPTY. SHEET: {}'.format(sheet_title))

    # Read column headers until end or 2 consecutive skipped headers
    for header in headers:
        # LOGGER.info('header = {}'.format(json.dumps(header, indent=2, sort_keys=True)))
        column_index = i + 1
        column_letter = colnum_string(column_index)
        header_value = header.get('formattedValue')
        if header_value: # if the column is NOT to be skipped
            column_is_skipped = False
            skipped = 0
            column_name = '{}'.format(header_value)
            if column_name in header_list:
                raise Exception('DUPLICATE HEADER ERROR: SHEET: {}, COL: {}, CELL: {}1'.format(
                    sheet_title, column_name, column_letter))
            header_list.append(column_name)

            first_value = None
            try:
                first_value = first_values[i]
            except IndexError as err:
                LOGGER.info('NO VALUE IN 2ND ROW FOR HEADER. SHEET: {}, COL: {}, CELL: {}2. {}'.format(
                    sheet_title, column_name, column_letter, err))
                first_value = {}
                first_values.append(first_value)
                pass

            column_effective_value = first_value.get('effectiveValue', {})

            col_val = None
            if column_effective_value == {}:
                if ("numberFormat" in first_value.get('effectiveFormat', {})):
                    column_effective_value_type = "numberValue"
                else:
                    column_effective_value_type = 'stringValue'
                    LOGGER.info('WARNING: NO VALUE IN 2ND ROW FOR HEADER. SHEET: {}, COL: {}, CELL: {}2.'.format(
                        sheet_title, column_name, column_letter))
                    LOGGER.info('   Setting column datatype to STRING')
            else:
                for key, val in column_effective_value.items():
                    if key in ('numberValue', 'stringValue', 'boolValue'):
                        column_effective_value_type = key
                        col_val = str(val)
                    elif key in ('errorType', 'formulaType'):
                        col_val = str(val)
                        raise Exception('DATA TYPE ERROR 2ND ROW VALUE: SHEET: {}, COL: {}, CELL: {}2, TYPE: {}'.format(
                            sheet_title, column_name, column_letter, key))

            column_number_format = first_values[i].get('effectiveFormat', {}).get(
                'numberFormat', {})
            column_number_format_type = column_number_format.get('type')

            # Determine datatype for sheet_json_schema
            #
            # column_effective_value_type = numberValue, stringValue, boolValue;
            #  INVALID: errorType, formulaType
            #  https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/other#ExtendedValue
            #
            # column_number_format_type = UNEPECIFIED, TEXT, NUMBER, PERCENT, CURRENCY,
            #   TIME, DATE_TIME, SCIENTIFIC
            #  https://developers.google.com/sheets/api/reference/rest/v4/spreadsheets/cells#NumberFormatType
            #
            column_format = None # Default

            if column_effective_value_type == 'stringValue':
                col_properties = {'type': ['null', 'string']}
                column_gs_type = 'stringValue'
            elif column_effective_value_type == 'boolValue':
                col_properties = {'type': ['null', 'boolean', 'string']}
                column_gs_type = 'boolValue'
            elif column_effective_value_type == 'numberValue':
                if column_number_format_type in ['DATE_TIME', 'DATE']:
                    col_properties = {
                        'type': ['null', 'string'],
                        'format': 'date-time'
                    }
                    column_gs_type = 'numberType.DATE_TIME'
                elif column_number_format_type == 'TIME':
                    col_properties = {
                        'type': ['null', 'string'],
                        'format': 'time'
                    }
                    column_gs_type = 'numberType.TIME'
                elif column_number_format_type == 'TEXT':
                    col_properties = {'type': ['null', 'string']}
                    column_gs_type = 'stringValue'
                elif column_number_format_type == 'CURRENCY':
                    col_properties = {'type': ['null', 'string']}
                    column_gs_type = 'stringValue'
                else:
                    # Interesting - order in the anyOf makes a difference.
                    # Number w/ singer.decimal must be listed last, otherwise errors occur.
                    col_properties = {
                        'type': ['null', 'string'],
                        'format': 'singer.decimal'
                    }
                    column_gs_type = 'numberType'
            # Catch-all to deal with other types and set to string
            # column_effective_value_type: formulaValue, errorValue, or other
            else:
                col_properties = {'type': ['null', 'string']}
                column_gs_type = 'unsupportedValue'
                LOGGER.info('WARNING: UNSUPPORTED 2ND ROW VALUE: SHEET: {}, COL: {}, CELL: {}2, TYPE: {}'.format(
                        sheet_title, column_name, column_letter, column_effective_value_type))
                LOGGER.info('Converting to string.')
        else: # if the column is to be skipped
            column_is_skipped = True
            skipped = skipped + 1
            column_index_str = str(column_index).zfill(2)
            column_name = '__sdc_skip_col_{}'.format(column_index_str)
            # unsupported field description if the field is to be skipped
            col_properties = {'type': ['null', 'string'], 'description': 'Column is unsupported and would be skipped because header is not available'}
            column_gs_type = 'stringValue'
            LOGGER.info('WARNING: SKIPPED COLUMN; NO COLUMN HEADER. SHEET: {}, COL: {}, CELL: {}1'.format(
                sheet_title, column_name, column_letter))
            LOGGER.info('  This column will be skipped during data loading.')

        if skipped >= 2:
            # skipped = 2 consecutive skipped headers
            # Remove prior_header column_name
            # stop scanning the sheet and break
            sheet_json_schema['properties'].pop(prior_header, None)
            # prior index is the index of the column prior to the currently column
            prior_index = column_index - 1
            # added a new boolean key `prior_column_skipped` to check if the column is one of the two columns with consecutive headers 
            # as due to consecutive empty headers both the columns should not be included in the schema as well as the metadata
            columns[prior_index-1]['prior_column_skipped'] = True
            LOGGER.info('TWO CONSECUTIVE SKIPPED COLUMNS. STOPPING SCAN AT: SHEET: {}, COL: {}, CELL {}1'.format(
                sheet_title, column_name, column_letter))
            break

        else:
            # skipped < 2 prepare `columns` dictionary with index, letter, column name, column type and 
            # if the column is to be skipped or not for each column in the list
            column = {}
            column = {
                'columnIndex': column_index,
                'columnLetter': column_letter,
                'columnName': column_name,
                'columnType': column_gs_type,
                'columnSkipped': column_is_skipped
            }
            columns.append(column)

            if column_gs_type in {'numberType.DATE_TIME', 'numberType.TIME', 'numberType'}:
                col_properties = {
                    'anyOf': [
                        col_properties,
                        {'type': ['null', 'string']} # all the time has string types in schema
                    ]
                }
            # add the column properties in the `properties` in json schema for the respective column name
            sheet_json_schema['properties'][column_name] = col_properties

        prior_header = column_name
        i = i + 1

    return sheet_json_schema, columns


# Get Header Row and 1st data row (Rows 1 & 2) from a Sheet on Spreadsheet w/ sheet_metadata query
#   endpoint: spreadsheets/{spreadsheet_id}
#   params: includeGridData = true, ranges = '{sheet_title}'!1:2
# This endpoint includes detailed metadata about each cell - incl. data type, formatting, etc.
def get_sheet_metadata(sheet, spreadsheet_id, client):
    sheet_id = sheet.get('properties', {}).get('sheetId')
    sheet_title = sheet.get('properties', {}).get('title')
    LOGGER.info('sheet_id = {}, sheet_title = {}'.format(sheet_id, sheet_title))

    stream_name = 'sheet_metadata'
    stream_obj = STREAMS.get(stream_name)(client, spreadsheet_id)
    api = stream_obj.api
    sheet_title_encoded = urllib.parse.quote_plus(sheet_title)
    sheet_title_escaped = re.escape(sheet_title)
    path, _ = stream_obj.get_path(sheet_title_encoded)

    sheet_md_results = client.get(path=path, api=api, endpoint=sheet_title_escaped)
    # sheet_metadata: 1st `sheets` node in results
    sheet_metadata = sheet_md_results.get('sheets')[0]

    # Create sheet_json_schema (for discovery/catalog) and columns (for sheet_metadata results)
    try:
        sheet_json_schema, columns = get_sheet_schema_columns(sheet_metadata)
    except Exception as err:
        LOGGER.warning('{}'.format(err))
        LOGGER.warning('SKIPPING Malformed sheet: {}'.format(sheet_title))
        sheet_json_schema, columns = None, None

    return sheet_json_schema, columns



from singer.catalog import Catalog, CatalogEntry, Schema
from tap_google_sheets.schema import STREAMS


def discover(client, spreadsheet_id):
    catalog = Catalog([])

    for stream, stream_obj in STREAMS.items():
        stream_object = stream_obj(client, spreadsheet_id)
        schemas, field_metadata = stream_object.get_schemas()

        # loop over the schema and prepare catalog
        for stream_name, schema_dict in schemas.items():

            schema = Schema.from_dict(schema_dict)
            mdata = field_metadata[stream_name]

            # get the primary keys for the stream
            #   if the stream is from STREAM, then get the key_properties
            #   else use the "table-key-properties" from the metadata
            if not STREAMS.get(stream_name):
                key_props = None
                # get primary key for the stream
                for mdt in mdata:
                    table_key_properties = mdt.get('metadata', {}).get('table-key-properties')
                    if table_key_properties:
                        key_props = table_key_properties
            else:
                stream_obj = STREAMS.get(stream_name)(client, spreadsheet_id)
                key_props = stream_obj.key_properties

            catalog.streams.append(CatalogEntry(
                stream=stream_name,
                tap_stream_id=stream_name,
                key_properties=key_props,
                schema=schema,
                metadata=mdata
            ))

    return catalog



import singer
from tap_google_sheets.streams import STREAMS, SheetsLoadData, write_bookmark, strftime

LOGGER = singer.get_logger()

def sync(client, config, catalog, state):
    """
    Sync the streams, loop over STREAMS
        "file_metadata" -> get the file's metadata and if the spreadsheet file is updated then continue the sync else stop the sync
        "spreadsheet_metadata" -> get the spreadsheet's metadata
            - sync the spreadsheet_metadata stream if selected
            - get the sheets in the spreadsheet and loop over the sheets and sync the sheet's records if selected
                - create 2 lists containing the data related the sheet's metadata and sheets loaded/synced during the sync
        "sheets_loaded" & "sheet_metadata" -> get the data lists from the "spreadsheet_metadata" stream and sync the records if selected
    """
    last_stream = singer.get_currently_syncing(state)
    LOGGER.info("last/currently syncing stream: %s", last_stream)

    selected_streams = []
    for stream in catalog.get_selected_streams(state):
        selected_streams.append(stream.stream)
    LOGGER.info("selected_streams: %s", selected_streams)

    if not selected_streams:
        # return if no stream is selected
        LOGGER.info("No stream is selected.")
        return

    # loop through main streams
    for stream_name, stream_obj in STREAMS.items():

        # get the stream object
        stream_obj = stream_obj(client, config.get("spreadsheet_id"), config.get("start_date"))

        # to sync the sheet's data, we need to get "spreadsheet_metadata"
        if stream_name == "spreadsheet_metadata":
            # get the metadata for the whole spreadsheet
            spreadsheet_metadata, time_extracted = stream_obj.get_data(stream_name=stream_obj.stream_name)

            # if the "spreadsheet_metadata" is selected, then do sync
            if stream_name in selected_streams:
                stream_obj.sync(catalog, state, spreadsheet_metadata, time_extracted)

            # get sheets from the metadata
            sheets = spreadsheet_metadata.get("sheets")
            # class to load sheet's data
            sheets_load_data = SheetsLoadData(client, config.get("spreadsheet_id"), config.get("start_date"))

            # perform sheet's sync and get sheet's metadata and sheet loaded records for "sheet_metadata" and "sheets_loaded" streams
            sheet_metadata_records, sheets_loaded_records = sheets_load_data.load_data(catalog=catalog,
                                                                                        state=state,
                                                                                        selected_streams=selected_streams,
                                                                                        sheets=sheets,
                                                                                        spreadsheet_time_extracted=time_extracted)

        # sync "sheet_metadata" and "sheets_loaded" based on the records from spreadsheet metadata
        elif stream_name in ["sheet_metadata", "sheets_loaded"] and stream_name in selected_streams:
            if stream_name == "sheet_metadata":
                stream_obj.sync(catalog, state, sheet_metadata_records)
            else:
                stream_obj.sync(catalog, state, sheets_loaded_records)

        # sync file metadata
        elif stream_name == "file_metadata":
            file_changed, file_modified_time = stream_obj.sync(catalog, state, selected_streams)
            if not file_changed:
                break

        LOGGER.info("FINISHED Syncing: %s", stream_name)

    # write "file_metadata" bookmark, as we have successfully synced all the sheet's records
    # it will force to re-sync of there is any interrupt between the sync
    write_bookmark(state, 'file_metadata', strftime(file_modified_time))



import math
import json
from datetime import datetime, timedelta
import pytz
import singer
from singer.utils import strftime

LOGGER = singer.get_logger()

# Tranform spreadsheet_metadata: add spreadsheetId, sheetUrl, and columns metadata
def transform_sheet_metadata(spreadsheet_id, sheet, columns):
    # Convert to properties to dict
    sheet_metadata = sheet.get('properties')
    sheet_metadata_tf = json.loads(json.dumps(sheet_metadata))
    sheet_id = sheet_metadata_tf.get('sheetId')
    sheet_url = 'https://docs.google.com/spreadsheets/d/{}/edit#gid={}'.format(
        spreadsheet_id, sheet_id)
    sheet_metadata_tf['spreadsheetId'] = spreadsheet_id
    sheet_metadata_tf['sheetUrl'] = sheet_url
    sheet_metadata_tf['columns'] = columns
    return sheet_metadata_tf

# Tranform spreadsheet_metadata: remove defaultFormat and sheets nodes, format as array
def transform_spreadsheet_metadata(spreadsheet_metadata):
    # Convert to dict
    spreadsheet_metadata_tf = json.loads(json.dumps(spreadsheet_metadata))
    # Remove keys: defaultFormat and sheets (sheets will come in sheet_metadata)
    if spreadsheet_metadata_tf.get('properties'):
        spreadsheet_metadata_tf['properties'].pop('defaultFormat', None)
    spreadsheet_metadata_tf.pop('sheets', None)
    # Add record to an array of 1
    spreadsheet_metadata_arr = []
    spreadsheet_metadata_arr.append(spreadsheet_metadata_tf)
    return spreadsheet_metadata_arr

# Tranform file_metadata: remove nodes from lastModifyingUser, format as array
def transform_file_metadata(file_metadata):
    # Convert to dict
    file_metadata_tf = json.loads(json.dumps(file_metadata))
    # Remove keys
    if file_metadata_tf.get('lastModifyingUser'):
        file_metadata_tf['lastModifyingUser'].pop('photoLink', None)
        file_metadata_tf['lastModifyingUser'].pop('me', None)
        file_metadata_tf['lastModifyingUser'].pop('permissionId', None)
    # Add record to an array of 1
    file_metadata_arr = []
    file_metadata_arr.append(file_metadata_tf)
    return file_metadata_arr

# Convert Excel Date Serial Number (excel_date_sn) to datetime string
# timezone_str: defaults to UTC (which we assume is the timezone for ALL datetimes)
def excel_to_dttm_str(string_value, excel_date_sn, timezone_str=None):
    if not timezone_str:
        timezone_str = 'UTC'
    tzn = pytz.timezone(timezone_str)
    sec_per_day = 86400
    excel_epoch = 25569 # 1970-01-01T00:00:00Z, Lotus Notes Serial Number for Epoch Start Date
    epoch_sec = math.floor((excel_date_sn - excel_epoch) * sec_per_day)
    epoch_dttm = datetime(1970, 1, 1)
    # For out of range values, it will throw OverflowError and it would return the string value
    # as passed in the sheets without any conversion
    try:
        excel_dttm = epoch_dttm + timedelta(seconds=epoch_sec)
    except OverflowError:
        return str(string_value), True
    utc_dttm = tzn.localize(excel_dttm).astimezone(pytz.utc)
    utc_dttm_str = strftime(utc_dttm)
    return utc_dttm_str, False


# transform datetime values in the sheet
def transform_sheet_datetime_data(value, unformatted_value, sheet_title, col_name, col_letter, row_num, col_type):
    if isinstance(unformatted_value, (int, float)):
        # passing both the formatted as well as the unformatted value, so we can use the string value in
        # case of any errors while datetime transform
        datetime_str, _ = excel_to_dttm_str(value, unformatted_value)
        return datetime_str
    else:
        LOGGER.info('WARNING: POSSIBLE DATA TYPE ERROR; SHEET: {}, COL: {}, CELL: {}{}, TYPE: {}'.format(
            sheet_title, col_name, col_letter, row_num, col_type))
        return str(value)

# transform time values in the sheet
def transform_sheet_time_data(value, unformatted_value, sheet_title, col_name, col_letter, row_num, col_type):
    if isinstance(unformatted_value, (int, float)):
        try:
            total_secs = unformatted_value * 86400 # seconds in day
            # Create string formatted like HH:MM:SS
            col_val = str(timedelta(seconds=total_secs))
        except ValueError:
            col_val = str(value)
            LOGGER.info('WARNING: POSSIBLE DATA TYPE ERROR; SHEET: {}, COL: {}, CELL: {}{}, TYPE: {}'.format(
                sheet_title, col_name, col_letter, row_num, col_type))
        return col_val
    else:
        return str(value)

# transform boolean values in the sheet
def transform_sheet_boolean_data(value, unformatted_value, sheet_title, col_name, col_letter, col_type, row):
    if isinstance(value, bool):
        return unformatted_value
    elif isinstance(value, str):
        if value.lower() in ('true', 't', 'yes', 'y'):
            col_val = True
        elif value.lower() in ('false', 'f', 'no', 'n'):
            col_val = False
        # As the float and the int values would be now returned as string itself, we need to check for the special
        # values as a string match rather than the integer/float match
        elif value in ('1', '-1', '1.00', '-1.00'):
            col_val = True
        elif value in ('0', '0.00'):
            col_val = False
        else:
            col_val = str(value)
            LOGGER.info('WARNING: POSSIBLE DATA TYPE ERROR; SHEET: {}, COL: {}, CELL: {}{}, TYPE: {}'.format(
                sheet_title, col_name, col_letter, row, col_type))
        return col_val
    elif isinstance(value, int):
        if value in (1, -1):
            col_val = True
        elif value == 0:
            col_val = False
        else:
            col_val = str(value)
            LOGGER.info('WARNING: POSSIBLE DATA TYPE ERROR; SHEET: {}, COL: {}, CELL: {}{}, TYPE: {}'.format(
                sheet_title, col_name, col_letter, row, col_type))
        return col_val
    elif isinstance(value, float):
        col_val = str(value)
        LOGGER.info('WARNING: POSSIBLE DATA TYPE ERROR; SHEET: {}, COL: {}, CELL: {}{}, TYPE: {}'.format(
            sheet_title, col_name, col_letter, row, col_type))
        return col_val

# transform decimal values in the sheet
def transform_sheet_decimal_data(formatted_value, unformatted_value, sheet_title, col_name, col_letter, row_num, col_type):
    """
        Transform number type data and return according to the datatype in the sheet

        :param
        formatted_value - The displayed value of a cell in the sheet ie. 2022-01-01
        unformatted_value - The calculated value of the field as per the value type ie. 44562,
            (the date values are converted into serial numbers by Google's API)
        sheet_title - The title of the sheet
        col_name - Column name
        col_letter - Column letter of the record ie. A, B, C, etc.
        row_num - Row number of the record
        col_type - Column type of the record (here: numberType)
    """

    # Removing comma to handle US number type format ie. 123,456.10 -> 123456.10
    numeric_value = formatted_value.replace(",", "")
    try:
        # Verify we can convert formatted value to float for scientific formatted numbers
        # For example:
        #   formatted value: "1.23E+03"
        #   unformatted value: 1234
        # thus, we can convert "1.23E+03" to float but, for int casting we get error and wrong value will be returned
        float(numeric_value)
    except ValueError:
        LOGGER.info('Received the value for sheet: {}, column: {}, cell: {}{} with unexpected data type. Ingesting this value with string format in the target'.format(
            sheet_title, col_name, col_letter, row_num
        ))
        return str(formatted_value) # Return original value in case of ValueError

    if type(unformatted_value) == int:
        return unformatted_value

    # For float type data, round off to 15 decimal digits
    # Determine float decimal digits
    decimal_digits = str(unformatted_value)[::-1].find('.')
    if decimal_digits > 15:
        try:
            # ROUND to multipleOf: 1e-15
            col_val = float(round(unformatted_value, 15))
        except ValueError:
            col_val = str(unformatted_value)
            LOGGER.info('WARNING: POSSIBLE DATA TYPE ERROR; SHEET: {}, COL: {}, CELL: {}{}, TYPE: {}'.format(
                sheet_title, col_name, col_letter, row_num, col_type))
        return col_val
    else: # decimal_digits <= 15, no rounding
        try:
            col_val = float(unformatted_value)
        except ValueError:
            col_val = str(unformatted_value)
            LOGGER.info('WARNING: POSSIBLE DATA TYPE ERROR: SHEET: {}, COL: {}, CELL: {}{}, TYPE: {}'.format(
                sheet_title, col_name, col_letter, row_num, col_type))
        return col_val

# transform number values in the sheet
def transform_sheet_number_data(formatted_value, unformatted_value, sheet_title, col_name, col_letter, row_num, col_type):
    """
        Handled number type records by type-casting into a float to verify the user has entered number type data.

        :param
        formatted_value - The displayed value of a cell in the sheet
        unformatted_value - The formatted value of a cell in the sheet
        sheet_title - The title of the sheet
        col_name - Column name
        col_letter - Column letter of the record ie. A, B, C, etc.
        row_num - Row number of the record
        col_type - Column type of the record (here: numberType)
    """
    if type(unformatted_value) in [int, float]:
        return transform_sheet_decimal_data(formatted_value, unformatted_value, sheet_title, col_name, col_letter, row_num, col_type)
    else:
        LOGGER.info('WARNING: POSSIBLE DATA TYPE ERROR: SHEET: {}, COL: {}, CELL: {}{}, TYPE: {} '.format(
                sheet_title, col_name, col_letter, row_num, col_type))
        return str(unformatted_value)

# return transformed column the values based on the datatype
def get_column_value(value, unformatted_value, sheet_title, col_name, col_letter, row_num, col_type, row):
        # NULL values
    if value is None or value == '':
        return None

    # Convert dates/times from Lotus Notes Serial Numbers
    # DATE-TIME
    elif col_type == 'numberType.DATE_TIME':
        return transform_sheet_datetime_data(value, unformatted_value, sheet_title, col_name, col_letter, row_num, col_type)

    # TIME ONLY (NO DATE)
    elif col_type == 'numberType.TIME':
        return transform_sheet_time_data(value, unformatted_value, sheet_title, col_name, col_letter, row_num, col_type)

    # NUMBER (INTEGER AND FLOAT)
    elif col_type == 'numberType':
        return transform_sheet_number_data(value, unformatted_value, sheet_title, col_name, col_letter, row_num, col_type)

    # STRING
    elif col_type == 'stringValue':
        return str(value)

    # BOOLEAN
    elif col_type == 'boolValue':
        return transform_sheet_boolean_data(value, unformatted_value, sheet_title, col_name, col_letter, col_type, row)

    # OTHER: Convert everything else to a string
    else:
        LOGGER.info('WARNING: POSSIBLE DATA TYPE ERROR; SHEET: {}, COL: {}, CELL: {}{}, TYPE: {}'.format(
            sheet_title, col_name, col_letter, row, col_type))
        return str(value)

# Transform sheet_data: add spreadsheet_id, sheet_id, and row, convert dates/times
#  Convert from array of values to JSON with column names as keys
def transform_sheet_data(spreadsheet_id, sheet_id, sheet_title, from_row, columns, sheet_data_rows, unformatted_rows):
    sheet_data_tf = []
    row_num = from_row
    # Create sorted list of columns based on columnIndex
    cols = sorted(columns, key=lambda i: i['columnIndex'])

    for (row, unformatted_row) in zip(sheet_data_rows, unformatted_rows):
        # If empty row, SKIP
        if row == []:
            LOGGER.info('EMPTY ROW: {}, SKIPPING'.format(row_num))
        else:
            sheet_data_row_tf = {}
            # Add spreadsheet_id, sheet_id, and row
            sheet_data_row_tf['__sdc_spreadsheet_id'] = spreadsheet_id
            sheet_data_row_tf['__sdc_sheet_id'] = sheet_id
            sheet_data_row_tf['__sdc_row'] = row_num
            col_num = 1
            for (value, unformatted_value) in zip(row, unformatted_row):
                # Select column metadata based on column index
                col = cols[col_num - 1]
                col_skipped = col.get('columnSkipped')
                if not col_skipped:
                    # Get column metadata
                    col_name = col.get('columnName')
                    col_type = col.get('columnType')
                    col_letter = col.get('columnLetter')

                    # get column value based on the type of the value
                    col_val = get_column_value(value, unformatted_value, sheet_title, col_name, col_letter, row_num, col_type, row)

                    sheet_data_row_tf[col_name] = col_val
                col_num = col_num + 1
            # APPEND non-empty row
            sheet_data_tf.append(sheet_data_row_tf)
        row_num = row_num + 1
    return sheet_data_tf, row_num



import json
import os
import time
import re
import simplejson as json
from collections import OrderedDict
import urllib.parse
import singer
import decimal
from singer import metrics, metadata, Transformer, utils, messages
from singer.utils import strptime_to_utc, strftime
from singer.messages import RecordMessage
from singer.transform import SchemaKey
import tap_google_sheets.transform as internal_transform
import tap_google_sheets.schema as schema

LOGGER = singer.get_logger()

def update_currently_syncing(state, stream_name):
    """
    Currently syncing sets the stream currently being delivered in the state.
    If the integration is interrupted, this state property is used to identify
        the starting point to continue from.
    Reference: https://github.com/singer-io/singer-python/blob/master/singer/bookmarks.py#L41-L46
    """
    if (stream_name is None) and ("currently_syncing" in state):
        del state["currently_syncing"]
    else:
        singer.set_currently_syncing(state, stream_name)
    singer.write_state(state)

def write_schema(catalog, stream_name):
    """
    Write schema from the stream
    """
    stream = catalog.get_stream(stream_name)
    schema = stream.schema.to_dict()
    try:
        singer.write_schema(stream_name, schema, stream.key_properties)
        LOGGER.info('Writing schema for: {}'.format(stream_name))
    except OSError as err:
        LOGGER.info('OS Error writing schema for: {}'.format(stream_name))
        raise err

def write_record(stream_name, record, time_extracted, version=None):
    """
    Write records for the stream with extracted time
    """
    try:
        # use "write_message" if version is found as
        # "write_record" params does not contain "version"
        if version:
            singer.messages.write_message(
                RecordMessage(
                    stream=stream_name,
                    record=record,
                    version=version,
                    time_extracted=time_extracted))
        else:
            singer.messages.write_record(
                stream_name=stream_name,
                record=record,
                time_extracted=time_extracted)
    except OSError as err:
        LOGGER.info('OS Error writing record for: {}'.format(stream_name))
        raise err

def get_bookmark(state, stream, default):
    """
    Get bookmark for the stream
    """
    if (state is None) or ('bookmarks' not in state):
        return default
    return (
        state
        .get('bookmarks', {})
        .get(stream, default)
    )

def write_bookmark(state, stream, value):
    """
    Write bookmark for the stream
    """
    if 'bookmarks' not in state:
        state['bookmarks'] = {}
    state['bookmarks'][stream] = value
    LOGGER.info('Write state for stream: {}, value: {}'.format(stream, value))
    singer.write_state(state)

def get_abs_path(path):
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), path)

def get_selected_fields(catalog, stream_name):
    """
    Get the list of selected fields for a stream from the catalog
    """
    stream = catalog.get_stream(stream_name)
    mdata = metadata.to_map(stream.metadata)
    mdata_list = singer.metadata.to_list(mdata)
    selected_fields = []
    for entry in mdata_list:
        field = None
        try:
            field = entry['breadcrumb'][1]
            if entry.get('metadata', {}).get('selected', False):
                selected_fields.append(field)
        except IndexError:
            pass
    return selected_fields

def new_format_message(message):
    """To override the ensure_ascii param, overwitten this function"""
    return json.dumps(message.asdict(), ensure_ascii=False, use_decimal=True)

# To override the ensure_ascii param as while writing record the currency symbols were written as ascii values,
# overwitten this function of messages file of the singer module
messages.format_message = new_format_message

class GoogleSheets:
    stream_name = None
    api = None
    path = None
    key_properties = None
    replication_method = None
    replication_keys = None
    params = None
    state = None

    def __init__(self, client, spreadsheet_id, start_date=None):
        self.client = client
        self.config_start_date = start_date
        self.spreadsheet_id = spreadsheet_id

    def get_path(self, sheet_title_encoded=""):
        """
        return path and query string for API Call
        """
        # Add in querystring parameters and replace {placeholder} variables
        # querystring function ensures parameters are added but not encoded causing API errors
        # create querystring for preparing the request
        querystring = '&'.join(['%s=%s' % (key, value) for (key, value) in self.params.items()]).replace('{sheet_title}', sheet_title_encoded)
        # create path for preparing the request
        path = '{}?{}'.format(self.path.replace('{spreadsheet_id}', self.spreadsheet_id), querystring)
        # return path and query string
        return path, querystring

    def get_schemas(self):
        """
        return schema for streams
        """
        schemas = {}
        field_metadata = {}

        schema_path = get_abs_path('schemas/{}.json'.format(self.stream_name))
        with open(schema_path) as file:
            schema = json.load(file)
        schemas[self.stream_name] = schema
        mdata = metadata.new()

        # Documentation:
        # https://github.com/singer-io/getting-started/blob/master/docs/DISCOVERY_MODE.md#singer-python-helper-functions
        # Reference:
        # https://github.com/singer-io/singer-python/blob/master/singer/metadata.py#L25-L44
        mdata = metadata.get_standard_metadata(
            schema=schema,
            key_properties=self.key_properties,
            valid_replication_keys=self.replication_keys,
            replication_method=self.replication_method
        )
        field_metadata[self.stream_name] = mdata

        return schemas, field_metadata

    def process_records(self, catalog, stream_name, records, time_extracted, version=None):
        """
        Transform/validate batch of records with schema and sent to target
        """
        stream = catalog.get_stream(stream_name)
        schema = stream.schema.to_dict()
        stream_metadata = metadata.to_map(stream.metadata)
        with metrics.record_counter(stream_name) as counter:
            for record in records:
                # Transform record for Singer.io
                with Transformer() as transformer:
                    try:
                        transformed_record = transformer.transform(
                            record,
                            schema,
                            stream_metadata)
                    except Exception as err:
                        LOGGER.error('{}'.format(err))
                        raise RuntimeError(err)
                    write_record(
                        stream_name=stream_name,
                        record=transformed_record,
                        time_extracted=time_extracted,
                        version=version)
                    counter.increment()
            return counter.value

    def get_data(self, stream_name, range_rows=None):
        """
        Call API for the steram and return response
        """
        if not range_rows:
            range_rows = ''
        # Replace {placeholder} variables in path
        # Encode stream_name: fixes issue w/ special characters in sheet name
        stream_name_escaped = re.escape(stream_name)
        stream_name_encoded = urllib.parse.quote_plus(stream_name)
        path = self.path.replace(
            '{spreadsheet_id}', self.spreadsheet_id).replace('{sheet_title}', stream_name_encoded).replace(
                '{range_rows}', range_rows)
        api = self.api
        _, querystring = self.get_path(stream_name_encoded)
        LOGGER.info('URL: {}/{}?{}'.format(self.client.base_url, path, querystring))
        data = {}
        time_extracted = utils.now()
        data = self.client.get(
            path=path,
            api=api,
            params=querystring,
            endpoint=stream_name_escaped)
        return data, time_extracted

    def sync_stream(self, records, catalog, time_extracted=None):
        """
        sync stream and write records
        """
        # Should sheets_loaded be synced?
        LOGGER.info('STARTED Syncing {}'.format(self.stream_name))
        update_currently_syncing(self.state, self.stream_name)
        selected_fields = get_selected_fields(catalog, self.stream_name)
        LOGGER.info('Stream: {}, selected_fields: {}'.format(self.stream_name, selected_fields))
        write_schema(catalog, self.stream_name)
        if not time_extracted:
            time_extracted = utils.now()
        record_count = self.process_records(
            catalog=catalog,
            stream_name=self.stream_name,
            records=records,
            time_extracted=time_extracted)
        LOGGER.info('FINISHED Syncing {}, Total Records: {}'.format(self.stream_name, record_count))
        update_currently_syncing(self.state, None)

class FileMetadata(GoogleSheets):
    stream_name = "file_metadata"
    api = "files"
    path = "files/{spreadsheet_id}"
    key_properties = ["id"]
    replication_method = "INCREMENTAL"
    replication_keys = ["modifiedTime"]
    params = {
        "fields": "id,name,createdTime,modifiedTime,version,teamDriveId,driveId,lastModifyingUser",
        "supportsAllDrives": True
    }

    def sync(self, catalog, state, selected_streams):
        """
        sync file's metadata
        """
        self.state = state
        # variable to check if file is changed or not

        # get date to start sync from, ie. start date or bookmark date
        start_date = strptime_to_utc(get_bookmark(state, self.stream_name, self.config_start_date))

        LOGGER.info("GET file_metadata")
        file_metadata, time_extracted = self.get_data(stream_name=self.stream_name)
        LOGGER.info("Transform file_metadata")

        file_modified_time = strptime_to_utc(file_metadata.get("modifiedTime"))
        LOGGER.info("last_datetime = {}, file_modified_time = {}".format(start_date, file_modified_time))
        if file_modified_time <= start_date:
            # if file is not changed, update the variable
            LOGGER.info("file_modified_time <= last_datetime, FILE NOT CHANGED. EXITING.")
            # return and stop syncing the next streams, as the file is not changed
            return False, file_modified_time

        # only perform sync if file metadata stream is selected and file is changed
        if self.stream_name in selected_streams:
            # transform file metadata records
            file_metadata_transformed = internal_transform.transform_file_metadata(file_metadata)
            # do sync
            self.sync_stream(file_metadata_transformed, catalog, time_extracted)

        return True, file_modified_time

class SpreadSheetMetadata(GoogleSheets):
    stream_name = "spreadsheet_metadata"
    api = "sheets"
    path = "spreadsheets/{spreadsheet_id}"
    key_properties = ["spreadsheetId"]
    replication_method = "FULL_TABLE"
    params = {
        "includeGridData": "false"
    }

    def get_schemas(self):
        """
        Get schema for spreadsheet and generate schema for the sheets in the spreadsheet
        """
        # get schema of spreadsheet metadata
        schemas, field_metadata = super().get_schemas()

        # prepare schema for sheets in the spreadsheet
        api = self.api
        path, querystring = self.get_path()

        # GET spreadsheet_metadata, which incl. sheets (basic metadata for each worksheet)
        spreadsheet_md_results = self.client.get(path=path, params=querystring, api=api, endpoint=self.stream_name)

        sheets = spreadsheet_md_results.get('sheets')
        if sheets:
            # Loop thru each worksheet in spreadsheet
            for sheet in sheets:
                # GET sheet_json_schema for each worksheet (from function above)
                sheet_json_schema, columns = schema.get_sheet_metadata(sheet, self.spreadsheet_id, self.client)

                # SKIP empty sheets (where sheet_json_schema and columns are None)
                if sheet_json_schema and columns:
                    sheet_title = sheet.get('properties', {}).get('title')
                    schemas[sheet_title] = sheet_json_schema
                    sheet_mdata = metadata.new()
                    sheet_mdata = metadata.get_standard_metadata(
                        schema=sheet_json_schema,
                        key_properties=['__sdc_row'],
                        valid_replication_keys=None,
                        replication_method='FULL_TABLE'
                    )
                    # for each column check if the `columnSkipped` value is true and the `prior_column_skipped` is false or None
                    # in the columns dict. The `prior_column_skipped` would be true  when it is the first column of the two
                    # consecutive empty headers column if true: update the incusion property to `unsupported`
                    for column in columns:
                        if column.get('columnSkipped') and not column.get('prior_column_skipped'):
                            mdata = metadata.to_map(sheet_mdata)
                            sheet_mdata = metadata.write(mdata, ('properties', column.get('columnName')), 'inclusion', 'unsupported')
                            sheet_mdata = metadata.to_list(mdata)
                    field_metadata[sheet_title] = sheet_mdata

        return schemas, field_metadata

    def sync(self, catalog, state, spreadsheet_metadata, time_extracted):
        """
        sync spreadsheet's metadata
        """
        self.state = state
        LOGGER.info('GET spreadsheet_metadata')

        # Transform spreadsheet_metadata
        LOGGER.info('Transform spreadsheet_metadata')
        spreadsheet_metadata_transformed = internal_transform.transform_spreadsheet_metadata(spreadsheet_metadata)

        # Sync spreadsheet_metadata if selected
        self.sync_stream(spreadsheet_metadata_transformed, catalog, time_extracted)

def new_transform(self, data, typ, schema, path):
    '''The new _transform function to replace in the singer module'''
    if self.pre_hook:
        data = self.pre_hook(data, typ, schema)

    if typ == "null":
        if data is None or data == "":
            return True, None
        else:
            return False, None

    elif schema.get("format") == "date-time":
        data = self._transform_datetime(data)
        if data is None:
            return False, None

        return True, data
    elif schema.get("format") == "singer.decimal":
        if data is None:
            return False, None

        if isinstance(data, (str, float, int)):
            try:
                return True, str(decimal.Decimal(str(data)))
            except:
                return False, None
        elif isinstance(data, decimal.Decimal):
            try:
                if data.is_snan():
                    return True, 'NaN'
                else:
                    return True, str(data)
            except:
                return False, None

        return False, None
    elif typ == "object":
        # Objects do not necessarily specify properties
        return self._transform_object(data,
                                        schema.get("properties", {}),
                                        path,
                                        schema.get(SchemaKey.pattern_properties))

    elif typ == "array":
        return self._transform_array(data, schema["items"], path)

    elif typ == "string":
        if data is not None:
            try:
                return True, str(data)
            except:
                return False, None
        else:
            return False, None

    elif typ == "integer":
        if isinstance(data, str):
            data = data.replace(",", "")

        try:
            return True, int(data)
        except:
            return False, None

    elif typ == "number":
        if isinstance(data, str):
            data = data.replace(",", "")
        try:
            return True, float(data)
        except:
            return False, None

    elif typ == "boolean":
        if data is None: # returns "null" if data is none
            return True, None
        if isinstance(data, str):  # return the data as string itself if the value is of type string
            return True, data
        try:
            return True, bool(data)
        except:
            return False, None

    else:
        return False, None

# To cast the boolean values differently, overwriting this function of Transformer class of 
# the singer module
Transformer._transform = new_transform

class SheetsLoadData(GoogleSheets):
    api = "sheets"
    path = "spreadsheets/{spreadsheet_id}/values/'{sheet_title}'!{range_rows}"
    data_key = "values"
    key_properties = ["spreadsheetId", "sheetId", "loadDate"]
    replication_method = "FULL_TABLE"
    params = {}

    def load_data(self, catalog, state, selected_streams, sheets, spreadsheet_time_extracted):
        """
        Load sheet's records if that sheet is selected for sync
        """
        self.state = state
        sheet_metadata = []
        sheets_loaded = []
        if sheets:
            # Loop through sheets (worksheet tabs) in spreadsheet
            for sheet in sheets:
                sheet_title = sheet.get('properties', {}).get('title')
                sheet_id = sheet.get('properties', {}).get('sheetId')

                # GET sheet_metadata and columns
                sheet_schema, columns = schema.get_sheet_metadata(sheet, self.spreadsheet_id, self.client)
                # LOGGER.info('sheet_schema: {}'.format(sheet_schema))

                # SKIP empty sheets (where sheet_schema and columns are None)
                if not sheet_schema or not columns:
                    LOGGER.info('SKIPPING Empty Sheet: {}'.format(sheet_title))
                else:
                    # Transform sheet_metadata
                    sheet_metadata_transformed = internal_transform.transform_sheet_metadata(self.spreadsheet_id, sheet, columns)
                    # LOGGER.info('sheet_metadata_transformed = {}'.format(sheet_metadata_transformed))
                    sheet_metadata.append(sheet_metadata_transformed)

                    # SHEET_DATA
                    # Should this worksheet tab be synced?
                    if sheet_title in selected_streams:
                        LOGGER.info('STARTED Syncing Sheet {}'.format(sheet_title))
                        update_currently_syncing(self.state, sheet_title)
                        selected_fields = get_selected_fields(catalog, sheet_title) # --------------------
                        LOGGER.info('Stream: {}, selected_fields: {}'.format(sheet_title, selected_fields))
                        write_schema(catalog, sheet_title)

                        # Emit a Singer ACTIVATE_VERSION message before initial sync (but not subsequent syncs)
                        # everytime after each sheet sync is complete.
                        # This forces hard deletes on the data downstream if fewer records are sent.
                        # https://github.com/singer-io/singer-python/blob/master/singer/messages.py#L137
                        last_integer = int(get_bookmark(self.state, sheet_title, 0))
                        activate_version = int(time.time() * 1000)
                        activate_version_message = singer.ActivateVersionMessage(
                                stream=sheet_title,
                                version=activate_version)
                        if last_integer == 0:
                            # initial load, send activate_version before AND after data sync
                            singer.write_message(activate_version_message)
                            LOGGER.info('INITIAL SYNC, Stream: {}, Activate Version: {}'.format(sheet_title, activate_version))

                        # Determine max range of columns and rows for "paging" through the data
                        sheet_last_col_index = 1
                        sheet_last_col_letter = 'A'
                        for col in columns:
                            col_index = col.get('columnIndex')
                            col_letter = col.get('columnLetter')
                            if col_index > sheet_last_col_index:
                                sheet_last_col_index = col_index
                                sheet_last_col_letter = col_letter
                        sheet_max_row = sheet.get('properties').get('gridProperties', {}).get('rowCount')

                        # Initialize paging for 1st batch
                        is_last_row = False
                        batch_rows = 200
                        from_row = 2
                        if sheet_max_row < batch_rows:
                            to_row = sheet_max_row
                        else:
                            to_row = batch_rows

                        # Loop thru batches (each having 200 rows of data)
                        while not is_last_row and from_row < sheet_max_row and to_row <= sheet_max_row:
                            range_rows = 'A{}:{}{}'.format(from_row, sheet_last_col_letter, to_row)

                            self.params = {
                                "dateTimeRenderOption": "SERIAL_NUMBER",
                                "valueRenderOption": "FORMATTED_VALUE",
                                "majorDimension": "ROWS"
                            }
                            # GET sheet_data for a worksheet tab
                            sheet_data, time_extracted = self.get_data(stream_name=sheet_title, range_rows=range_rows)
                            # Data is returned as a list of arrays, an array of values for each row
                            sheet_data_rows = sheet_data.get('values', [])
                            self.params = {
                                "dateTimeRenderOption": "SERIAL_NUMBER",
                                "valueRenderOption": "UNFORMATTED_VALUE",
                                "majorDimension": "ROWS"
                            }
                            unformatted_sheet_data, _ = self.get_data(stream_name=sheet_title, range_rows=range_rows)
                            unformatted_sheet_data_rows = unformatted_sheet_data.get('values', [])

                            # Transform batch of rows to JSON with keys for each column
                            sheet_data_transformed, row_num = internal_transform.transform_sheet_data(
                                spreadsheet_id=self.spreadsheet_id,
                                sheet_id=sheet_id,
                                sheet_title=sheet_title,
                                from_row=from_row,
                                columns=columns,
                                sheet_data_rows=sheet_data_rows, 
                                unformatted_rows = unformatted_sheet_data_rows)

                            # Here row_num is the addition of from_row and total records get in response(per batch).
                            # Condition row_num < to_row was checking that if records on the current page are less than expected(to_row) or not.
                            # If the condition returns true then it was breaking the loop.
                            # API does not return the last empty rows in response.
                            # For example, rows 199 and 200 are empty, and a total of 400 rows are there in the sheet. So, in 1st iteration,
                            # to_row = 200, from_row = 2, row_num = 2(from_row) + 197 = 199(1st row contain header value)
                            # So, the above condition become true and breaks the loop without syncing records from 201 to 400.
                            # sheet_data_rows is no of records return in the current page. If it's a whole blank page then stop looping.
                            # So, in the above case, it syncs records 201 to 400 also even if rows 199 and 200 are blank.
                            # Then when the next batch 401 to 600 is empty, it breaks the loop.
                            if not sheet_data_rows: # If a whole blank page found, then stop looping.
                                is_last_row = True

                            # Process records, send batch of records to target
                            record_count = self.process_records(
                                catalog=catalog,
                                stream_name=sheet_title,
                                records=sheet_data_transformed,
                                time_extracted=spreadsheet_time_extracted,
                                version=activate_version)
                            LOGGER.info('Sheet: {}, records processed: {}'.format(
                                sheet_title, record_count))

                            # Update paging from/to_row for next batch
                            from_row = to_row + 1
                            if to_row + batch_rows > sheet_max_row:
                                to_row = sheet_max_row
                            else:
                                to_row = to_row + batch_rows

                        # End of Stream: Send Activate Version and update State
                        singer.write_message(activate_version_message)
                        write_bookmark(self.state, sheet_title, activate_version)
                        LOGGER.info('COMPLETE SYNC, Stream: {}, Activate Version: {}'.format(sheet_title, activate_version))
                        LOGGER.info('FINISHED Syncing Sheet {}, Total Rows: {}'.format(
                            sheet_title, row_num - 2)) # subtract 1 for header row
                        update_currently_syncing(self.state, None)

                        # SHEETS_LOADED
                        # Add sheet to sheets_loaded
                        sheet_loaded = {}
                        sheet_loaded['spreadsheetId'] = self.spreadsheet_id
                        sheet_loaded['sheetId'] = sheet_id
                        sheet_loaded['title'] = sheet_title
                        sheet_loaded['loadDate'] = strftime(utils.now())
                        sheet_loaded['lastRowNumber'] = row_num
                        sheets_loaded.append(sheet_loaded)

        return sheet_metadata, sheets_loaded

class SheetMetadata(GoogleSheets):
    stream_name = "sheet_metadata"
    api = "sheets"
    path = "spreadsheets/{spreadsheet_id}"
    key_properties = ["sheetId"]
    replication_method = "FULL_TABLE"
    params = {
        "includeGridData": "true",
        "ranges": "'{sheet_title}'!1:2"
    }

    def sync(self, catalog, state, sheet_metadata_records):
        """
        Write sheet's metadata records
        """
        self.state = state
        self.sync_stream(sheet_metadata_records, catalog)

class SheetsLoaded(GoogleSheets):
    stream_name = "sheets_loaded"
    api = "sheets"
    path = "spreadsheets/{spreadsheet_id}/values/'{sheet_title}'!{range_rows}"
    data_key = "values"
    key_properties = ["spreadsheetId", "sheetId", "loadDate"]
    replication_method = "FULL_TABLE"
    params = {
        "dateTimeRenderOption": "SERIAL_NUMBER",
        "valueRenderOption": "UNFORMATTED_VALUE",
        "majorDimension": "ROWS"
    }

    def sync(self, catalog, state, sheets_loaded_records):
        """
        Write sheets loaded records
        """
        self.state = state
        self.sync_stream(sheets_loaded_records, catalog)


# create OrderDict, as the order matters for syncing the streams
# "file_metadata" -> do not sync other streams, if file is not changed
# "spreadsheet_metadata" -> get sheets in the spreadsheet and load sheet's records
#       and prepare records for "sheet_metadata" and "sheets_loaded" streams
STREAMS = OrderedDict()
STREAMS['file_metadata'] = FileMetadata
STREAMS['spreadsheet_metadata'] = SpreadSheetMetadata
STREAMS['sheet_metadata'] = SheetMetadata
STREAMS['sheets_loaded'] = SheetsLoaded
